from django.core.exceptions import ValidationError
from django.views.generic.list import MultipleObjectMixin

from .base import MapperMixin, NapView


# List views
class ListMixin(MapperMixin, MultipleObjectMixin):

    def ok_response(self, **kwargs):
        '''
        Shortcut to return a ``multiple_response`` with a ``status`` of
        ``self.ok_status``.
        '''
        kwargs.setdefault('status', self.ok_status)
        return self.multiple_response(**kwargs)


class ListGetMixin:

    def get(self, request, *args, **kwargs):
        '''
        Returns an ok_response
        '''
        return self.ok_response()


class ListPostMixin:

    def post(self, request, *args, **kwargs):
        '''
        Handle POST on a List view.

        Validates the data against the `model_class`, and calls `post_valid` or
        `post_invalid` as appropriate.
        '''
        self.mapper = self.get_mapper(self.model())
        self.data = self.get_request_data()

        try:
            self.object = self.mapper._apply(self.data)
        except ValidationError:
            return self.post_invalid(self.mapper._errors)

        return self.post_valid()

    def post_invalid(self, errors):
        '''
        Called on invalid POST data.

        Returns an `error_response`.
        '''
        return self.error_response(errors)

    def post_valid(self, **kwargs):
        '''
        Called on valid POST data.

        Saves self.object and returns a `created_response`.
        '''
        self.object.save()

        return self.created_response(**kwargs)


class ListBaseView(ListMixin, NapView):
    '''
    A Subclass of ListMixin and NapView.
    '''
    pass



from functools import update_wrapper

from . import BaseHttpResponse


class except_response:
    '''
    Allow view functions to handle HttpResponse as exception.
    '''
    def __init__(self, func):
        self.func = func
        update_wrapper(self, func)

    def __call__(self, request, *args, **kwargs):
        try:
            return self.func(request, *args, **kwargs)
        except BaseHttpResponse as resp:
            return resp



from django.core.exceptions import ValidationError
from django.views.generic.detail import SingleObjectMixin

from .base import MapperMixin, NapView


# Object views
class ObjectMixin(MapperMixin, SingleObjectMixin):

    def ok_response(self, **kwargs):
        kwargs.setdefault('status', self.ok_status)
        return self.single_response(**kwargs)


class ObjectGetMixin:

    def get(self, request, *args, **kwargs):
        return self.ok_response()


class ObjectPutMixin:

    def put(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.mapper = self.get_mapper(self.object)

        self.data = self.get_request_data({})

        try:
            self.mapper._apply(self.data)
        except ValidationError:
            return self.put_invalid(self.mapper._errors)

        return self.put_valid()

    def put_valid(self, **kwargs):
        self.object.save()
        return self.ok_response(**kwargs)

    def put_invalid(self, errors):
        return self.error_response(errors)


class ObjectPatchMixin:

    def patch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.mapper = self.get_mapper(self.object)

        self.data = self.get_request_data({})

        try:
            self.mapper._patch(self.data)
        except ValidationError:
            return self.patch_invalid(self.mapper._errors)

        return self.patch_valid()

    def patch_valid(self, **kwargs):
        self.object.save()
        return self.ok_response(**kwargs)

    def patch_invalid(self, errors):
        return self.error_response(errors)


class ObjectDeleteMixin:

    def delete(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.mapper = self.get_mapper(self.object)

        return self.delete_valid()

    def delete_valid(self, **kwargs):
        self.object.delete()

        return self.deleted_response(**kwargs)


class ObjectBaseView(ObjectMixin, NapView):
    pass



import json

import requests


class RPCProxy:
    def __init__(self, client, name):
        self.client = client
        self.name = name

    def __call__(self, **kwargs):
        resp = self.client.session.post(
            self.client.endpoint,
            data=json.dumps(kwargs),
            headers={
                'X-Rpc-Action': self.name,
                'Content-Type': 'application/json',
            },
        )
        return resp.json()


class RPCClient:
    def __init__(self, endpoint):
        self.endpoint = endpoint
        self.session = requests.Session()

    def __getattr__(self, key):
        return RPCProxy(self, key)



import json
from cgi import FieldStorage
from inspect import isgenerator

from django.core.serializers.json import DjangoJSONEncoder
from django.http import QueryDict

from .. import http

# JSONDecodeError was added in Python 3.5
try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


class JsonMixin:
    '''
    Common methods for handling JSON request data.
    '''
    CONTENT_TYPES = ['application/json', 'text/json']
    JSON_DECODER = None

    def get_request_data(self, default=None):
        '''Retrieve data from request'''
        from django.conf import settings
        encoding = self.request.encoding or settings.DEFAULT_CHARSET

        content_type = self.request.content_type

        if content_type in self.CONTENT_TYPES:
            if not self.request.body:
                return default
            try:
                return self.loads(self.request.body.decode(encoding))
            except JSONDecodeError:
                raise http.BadRequest()

        if self.request.method in ('PUT', 'PATCH'):
            if content_type == 'application/x-www-form-urlencoded':
                return QueryDict(self.request.body, encoding=encoding)
            elif content_type == 'multipart/form-data':
                return FieldStorage(
                    self.request,
                    headers=self.request.META,
                    strict_parsing=True,
                    encoding=encoding,
                )

        return self.request.POST

    def loads(self, data, **kwargs):
        '''Serialise data for responses.'''
        kwargs.setdefault('cls', self.JSON_DECODER)
        return json.loads(data, **kwargs)


class List(list):
    '''
    A list-alike for generators to fool json.dump.
    '''
    def __init__(self, gen):
        self.gen = gen

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.gen)

    def __bool__(self):
        return True


class NapJSONEncoder(DjangoJSONEncoder):

    def default(self, o):
        if isgenerator(o):
            return List(o)
        m = getattr(o, '__json__', None)
        if callable(m):
            return m()
        return super().default(o)



import datetime
from functools import partial

from django.core.exceptions import ValidationError
from django.db.models.fields import NOT_PROVIDED


class field(property):
    '''A base class to compare against.'''
    def __new__(cls, *args, **kwargs):
        '''
        Allow specifying keyword arguments when used as a decorator.
        '''
        if not args:
            return partial(cls, **kwargs)
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, fget=None, fset=None, fdel=None, doc=None, **kwargs):
        self.required = kwargs.pop('required', False)
        self.default = kwargs.pop('default', NOT_PROVIDED)
        self.readonly = kwargs.pop('readonly', fset is None)
        self.null = kwargs.pop('null', False)
        super().__init__(fget, fset, fdel, doc)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return self.fget(instance._obj)

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(instance._obj, value)

    def setter(self, func):
        kwargs = self.__dict__.copy()
        kwargs.pop('__doc__')
        kwargs.pop('readonly')
        return self.__class__(self.fget, func, **kwargs)


class context_field(field):
    '''Special case of field that allows access to the Mapper itself'''
    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return self.fget(instance._obj, instance._context)

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        return self.fset(instance._obj, value, instance._context)


class Field(field):
    def __init__(self, attr, *args, **kwargs):
        self.attr = attr
        kwargs.setdefault('readonly', False)
        super().__init__(*args, **kwargs)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        value = getattr(instance._obj, self.attr, self.default)
        if self.null and value is None:
            return value
        return self.get(value)

    def __set__(self, instance, value):
        if self.readonly:
            raise AttributeError('Field is read-only.')
        if value is None:
            if not self.null:
                raise ValueError('Field may not be None')
        else:
            try:
                value = self.set(value)
            except (TypeError, ValueError):
                raise ValidationError('Invalid value')
        setattr(instance._obj, self.attr, value)

    def get(self, value):
        return value

    def set(self, value):
        return value


class BooleanField(Field):
    def set(self, value):
        if isinstance(value, bool):
            return value
        return value.lower() in (1, '1', 't', 'y', 'true')


class IntegerField(Field):
    def get(self, value):
        return int(value)

    def set(self, value):
        return int(value)


class FloatField(Field):
    def get(self, value):
        return float(value)

    def set(self, value):
        return float(value)


class TimeField(Field):
    def get(self, value):
        return value.replace(microsecond=0).isoformat()

    def set(self, value):
        if isinstance(value, datetime.time):
            return value
        return datetime.datetime.strptime(value, '%H:%M:%S').time()


class DateField(Field):
    def get(self, value):
        return value.isoformat()

    def set(self, value):
        if isinstance(value, datetime.date):
            return value
        return datetime.datetime.strptime(value, '%Y-%m-%d').date()


class DateTimeField(Field):
    def get(self, value):
        return value.replace(microsecond=0).isoformat(' ')

    def set(self, value):
        if isinstance(value, datetime.datetime):
            return value
        return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')


class MapperField(Field):
    '''
    A field that passes data through a Mapper.
    '''
    def __init__(self, *args, **kwargs):
        self.mapper = kwargs.pop('mapper')
        super().__init__(*args, **kwargs)

    def get(self, value):
        return self.mapper() << value

    def set(self, value):
        return value >> self.mapper()



from setuptools import setup, find_packages

with open('README.rst') as fin:
    readme = fin.read()

setup(
    name='django-nap',
    version='0.40.0',
    description='A light REST tool for Django',
    long_description=readme,
    author='Curtis Maloney',
    author_email='curtis@tinbrain.net',
    url='http://github.com/funkybob/django-nap',
    keywords=['django', 'json', 'rest', 'api'],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    zip_safe=False,
    classifiers = [
        'Environment :: Web Environment',
        'Framework :: Django',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: Implementation :: PyPy',
    ],
    requires = [
        'Django (>=2.0)',
    ],
    extras_require={
        'rps': ['requets'],
    },
    install_requires = [
        'Django>=2.0',
    ],
)



from .base import NapView  # NOQA
from .list import ListMixin, ListGetMixin, ListPostMixin, ListPostMixin, ListBaseView  # NOQA
from .object import ObjectMixin, ObjectGetMixin, ObjectPutMixin, ObjectPatchMixin, ObjectDeleteMixin, ObjectBaseView  # NOQA



from django.views.generic import View

from nap import http
from nap.http.decorators import except_response
from nap.utils import JsonMixin


class NapView(View):
    '''
    Base view for Nap CBV.

    Catches any http exceptions raised, and returns them instead.
    '''

    @classmethod
    def as_view(cls, **initkwargs):
        '''
        If any code raises one of our HTTP responses, we should catch and
        return it.
        '''
        return except_response(super().as_view(**initkwargs))


class MapperMixin(JsonMixin):
    '''
    Base class for generating JSON responses using Mappers.
    '''
    response_class = http.JsonResponse
    content_type = 'application/json'
    mapper_class = None
    include_meta = False

    ok_status = http.STATUS.OK
    accepted_status = http.STATUS.ACCEPTED
    created_status = http.STATUS.CREATED
    deleted_status = http.STATUS.NO_CONTENT
    error_status = http.STATUS.BAD_REQUEST

    def get_mapper(self, obj=None):
        '''
        Get the mapper to use for this request.

        Default action is to use self.mapper_class
        '''
        return self.mapper_class(obj)

    def empty_response(self, **kwargs):
        '''
        Helper method to return an empty response.
        '''
        kwargs.setdefault('safe', False)
        return self.response_class('', **kwargs)

    def single_response(self, **kwargs):
        '''
        Helper method to return a single object.

        If `object` is not passed, it will try to use `self.object`.  If
        `self.object` is not set, it will call `self.get_object()`.

        If `mapper` is not passed, it will try to use `self.mapper`.  If
        `self.mapper` is not set, it will call `self.get_mapper()`.

        Returns a `self.response_class` instance, passed ``mapper << obj``,
        along with `**kwargs`.
        '''
        try:
            obj = kwargs.pop('object')
        except KeyError:
            try:
                obj = self.object
            except AttributeError:
                obj = self.get_object()

        try:
            mapper = kwargs.pop('mapper')
        except KeyError:
            try:
                mapper = self.mapper
            except AttributeError:
                mapper = self.get_mapper(obj)

        return self.response_class(mapper << obj, **kwargs)

    def multiple_response(self, **kwargs):
        '''
        Helper method to return an iterable of objects.

        If `object_list` is not passed, it will try to use `self.obect_list`.
        If `self.object_list` is not set, it will call
        `self.get_object_list()`.

        If `mapper` is not passed, it will try to use `self.mapper`.  If
        `self.mapper` is not set, it will call `self.get_mapper()`.

        Returns a `self.response_class` instance, passed a list of ``mapper <<
        obj`` applied to each object, along with `**kwargs`.
        '''
        kwargs.setdefault('safe', False)

        try:
            object_list = kwargs.pop('object_list')
        except KeyError:
            try:
                object_list = self.object_list
            except AttributeError:
                object_list = self.get_queryset()

        try:
            mapper = kwargs.pop('mapper')
        except KeyError:
            try:
                mapper = self.mapper
            except AttributeError:
                mapper = self.get_mapper()

        page_size = self.get_paginate_by(object_list)
        if page_size:
            paginator, page, object_list, is_paginated = self.paginate_queryset(object_list, page_size)
        else:
            paginator = page = None
            is_paginated = False

        data = [mapper << obj for obj in object_list]

        if page_size or self.include_meta:
            meta = self.get_meta(page)
            data = {
                'meta': meta,
                'data': data,
            }

        return self.response_class(data, **kwargs)

    def get_meta(self, page):
        if not page:
            return {}
        return {
            'offset': page.start_index() - 1,
            'page': page.number,
            'total': page.paginator.count,
        }

    def accepted_response(self, **kwargs):
        '''
        Shortcut to return an ``empty_response`` using a ``status`` of
        ``self.accepted_status``.
        '''
        kwargs.setdefault('status', self.accepted_status)
        return self.empty_response(**kwargs)

    def created_response(self, **kwargs):
        '''
        Shortcut to return a ``single_response`` using a ``status`` of
        ``self.created_status``.
        '''
        kwargs.setdefault('status', self.created_status)
        return self.single_response(**kwargs)

    def deleted_response(self, **kwargs):
        '''
        Shortcut to return an ``empty_response`` using a ``status`` of
        ``self.deleted_status``.
        '''
        kwargs.setdefault('status', self.deleted_status)
        return self.empty_response(**kwargs)

    def error_response(self, errors):
        '''
        Helper method to return ``self.respone_class`` with a ``status`` of
        ``self.error_status``.

        Will flatten ``errors`` using ``get_json_data``.
        '''
        return self.response_class(
            errors.get_json_data(),
            status=self.error_status,
        )




class Writer:
    '''
    A generator friendly, unicode aware CSV encoder class built for speed.

    >>> csv = Writer(fields=['a', 'b', 'c'])
    >>> csv.write_headers()
    b'a,b,c\n'
    >>> csv.write([1, '2,', 'c'])
    b'1,"2,",c\n'

    '''
    __slots__ = ('fields', 'headers', 'SEP', 'QUOTE', 'ESCQUOTE', 'LINEBREAK', 'ENCODING', 'escape_field')

    def __init__(self, fields,
                 headers=None,
                 SEP=u',',                  # What to put between fields
                 QUOTE=u'"',                # What to wrap fields in, if they contain SEP
                 ESCQUOTE=None,             # What to replace a QUOTE in a field with
                 LINEBREAK=u'\n',           # What to put between records
                 ENCODING='utf-8',
                 ):
        '''
        opts MUST contain 'fields', a list of field names.
        opts may also include 'headers', a list of field headings.
        opts MAY override any of the above configurables.
        '''
        self.fields = fields
        self.headers = headers or fields
        self.SEP = SEP
        self.QUOTE = QUOTE
        self.ESCQUOTE = ESCQUOTE or QUOTE + QUOTE
        self.LINEBREAK = LINEBREAK
        self.ENCODING = ENCODING

        def escape_field(val, SEP=self.SEP, QUOTE=self.QUOTE, ESCQ=self.ESCQUOTE):
            '''
            Escape separator and quote values, and wrap with quotes if needed

            Inlined for speed
            '''
            # escape quotes in the value
            val = str(val).replace(QUOTE, ESCQ)
            # if needed, wrap quotes around value
            if SEP in val or QUOTE in val:
                return QUOTE + val + QUOTE
            return val
        self.escape_field = escape_field

    def write(self, values):
        '''Write a row of values'''
        line = self.SEP.join(map(self.escape_field, values)) + self.LINEBREAK
        if self.ENCODING:
            line = line.encode(self.ENCODING)
        return line

    def write_dict(self, values):
        '''Write a row, getting values from a dict.'''
        return self.write(map(values.get, self.fields))

    def write_headers(self):
        '''Write a row of headers.'''
        return self.write(self.headers)
